/*
 *	   o p e n r l l _ c
 *
 *	This file is part of the LIBRLL library.
 *	It handles opening a RLL library.
 *
 * (c) Copyright 1999 David J. Walker
 *	   Email:  dave.walker@icl.com
 *
 *	   Permission to copy and/or distribute is granted under the
 *	   following conditions:
 *
 *	   1). This notice must remain intact.
 *	   2). The author is not responsible for the consequences of use
 *		   this software, no matter how awful, even if they arise
 *		   from defects in it.
 *	   3). Altered versions must not be represented as being the
 *		   original software.
 */

#include "librll.h"
#include <fcntl.h>


RLL_FILE *	RLL_OpenRLL (char * filename)
{
	int fd;

	if (0 <= (fd = open(filename, O_RDONLY)))
	{
		return NULL;
	}
	return RLL_OpenRLLFile (fd);
}


RLL_FILE *	RLL_OpenRllFile (int fd)
{
	char	buffer[256];
	RLL_FILE *	rp;
	short	thing_length;
	char	rlib_number;
	char	rlib_thing[256];
	char	rlib_version[5];
	BSS_HEADER	bsshead;
	RLL_FILE *	rp;

	size_t	UseRll;

	DBG_ENTER(("RLL_OPENRLLFILE",0x21,""));

	/*
	 *	Open file and allocate standard areas.
	 */
	if (NULL == (rp = _RLL_OpenMem(fd)))
	{
		return NULL;
	}

	/*
	 *	Read THING header
	 *	(& check flaga)
	 */

	(void) read (rp->fd, buffer, 38);
	DBG(("RLL_OPENRLLFILE", 0x18, "read THING version and name"));
	(void) read (rp->fd, rlib_version, 4);
	rlib_version[4] = '\0';

	(void) read (rp->fd, &thing_length, 2);
	(void) read (rp->fd, rlib_thing, thing_length);
	if (thing_length & 1)
	{
		(void) read (rp->fd, buffer, 1);
	}

	if (NULL == (rp->rll_thing = malloc (strlen(rlib_thing)))
	{
		return NULL;
	}
	if (NULL == (rp->rll_header = malloc(sizeof(RLL_HEADER))))
	{
		return NULL;
	}
	/*
	 *	Now read in RLL header
	 */

	DBG(("LINK_RLL", 0x18, "read THING tag"));
	(void) read (inp_hnd, buffer, 8);
	if (strcmp(buffer, "THG%"))
	{
		Halt (ERR_BINFILE);
	}
	DBG(("LINK_RLL", 0x18, "read RLL tag"));
	(void) read (inp_hnd, buffer, 8);
	if (strncmp(buffer, "<<RLL>>", 8))
	{
		Halt (ERR_BINFILE);
	}
	/*
	 *	Get offset of BSS header
	 */
	DBG(("LINK_RLL", 0x18, "read BSS Header offset"));
	(void) read (inp_hnd, &symoffset, 4);
	/*
	 *	Read BSS Header
	 *	(& check flag)
	 */
	DBG(("LINK_RLL", 0x18, "read BSS Header (at offset %ld)", symoffset));
	(void) lseek (inp_hnd, symoffset, SEEK_SET);
	(void) read (inp_hnd, &bsshead, sizeof(BSS_HEADER));
	if (strncmp(bsshead.bss_iden, BSSFLAG, sizeof (bsshead.bss_iden)))
	{
		Halt (ERR_BINFILE);
	}
	/*
	 *	Position at XDEF area
	 *	and then read through entries
	 *	setting up use where needed
	 */
	DBG(("LINK_RLL", 0x18, "Go to XDEF area at %ld", bsshead.bss_xdef));
	(void) lseek (inp_hnd, bsshead.bss_xdef, SEEK_SET);
	DBG(("LINK_RLL", 0x18, "read XDEF area tag"));
	(void) read (inp_hnd, buffer, 8);
	if (strncmp(buffer, BSS_XDEFFLAG, 8))
	{
		Halt (ERR_BINFILE);
	}
	/*
	 *	Read XDEF names and check if they are needed
	 */
	DBG(("LINK_FILE", 0x18, "now work through symbols in this RLL"));
	for (UseRll = 0, rlib_number = LIST_Count(rlib_list) + 1 ;	; )
	{
		XSYMBOL * xsy;

		DBG(("LINK_RLL", 0x18, "current file position=%ld", \
										lseek (inp_hnd, 0, SEEK_CUR)));
		if (sizeof(symlength) != read (inp_hnd, &symlength, sizeof(symlength))
		||	sizeof(symoffset) != read (inp_hnd, &symoffset, sizeof(symoffset)))
		{
			Halt (ERR_BINFILE);
		}
		DBG(("LINK_RLL", 0x18, "Symbol offset=%ld ($%p), length=%d", \
									symoffset, symoffset, symlength));
		if (0 == symlength)
		{
			DBG(("LINK_RLL", 0x11, "Exit - end of table reached"));
			break;
		}
		(void) read (inp_hnd, buffer, symlength);
		buffer[symlength] = '\0';
		DBG(("LINK_RLL", 0x18, "Next symbol is '%s'", buffer));
		if (NULL != (xsy = Symbol_Find(buffer)))
		{
			if (0 == (xsy->xsy_defd & XSYMBOL_DEFINED))
			{
				DBG(("LINK_RLL", 0x18, "Symbol %s used from this RLL", buffer));
				xsy->xsy_defd |= (XSYMBOL_DEFINED | XSYMBOL_RLL);
				xsy->xsy_rlib = rlib_number;
				UseRll++;
			}
			else
			{
				DBG(("LINK_RLL", 0x18, "Symbol not used - already defined"));
			}
		}
		else
		{
			DBG(("LINK_RLL", 0x18, "Symbol not used - not wanted"));
		}
	}
	DBG(("LINK_RLL", 0x18, "finished checking symbols in this RLL"));

	/*
	 *	If we used any XDEFs then
	 *	add this to the RLL's used.
	 */
	DBG(("LINK_RLL", 0x18, "%ld symbols used from this RLL", UseRll));
	if (0 != UseRll)
	{
		RLL_ENTRY * rll_entry;

		DBG(("LINK_RLL", 0x18, "FileName=%s, ThingName=%s, ThingVersion=%s", \
								((NAMELIST *)filenode)->name,
								rlib_thing, rlib_version));

		if (NULL == (rll_entry = LIST_NewNode (rlib_list)))
		{
			Halt (ERR_NOMEMORY);
		}
		rll_entry->filename =  NameSpace_Use (((NAMELIST *)filenode)->name);
		rll_entry->thingname = NameSpace_Use (rlib_thing);
		rll_entry->version = NameSpace_Use (rlib_version);
		(void) LIST_Append (rlib_list, rll_entry);
	}

	DBG_EXIT(("LINK_RLL",0x21,""));
	return;
}
	return rp;
}
